/* llm - litle lisp machine:
 * Author: Christian Barthel <bch@online.de>
 * Compile: flex tokenize.l
 *	        cc lex.yy.c -lfl
 */
%{
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <err.h>
struct token;
struct token {
  int           type;
	union {
    char       *str;
    int         num;
  } v;
  struct token *next;
};

struct ast; 
struct ast {
  int             type;
  union {
    struct token *token;
    struct ast   *list;
  } v;
  struct ast     *next;
};

struct env;
struct env {
  char       *name;
  struct ast *a;
  struct env *parent;
  struct env *next;
};

enum asttype {
  AST_TOK  = 400,
  AST_LIST = 401,
  AST_QUOTE = 402
};
enum yytokentype {
  NUM    = 258,
  LPAR   = 259,
  RPAR   = 260,
  SYM    = 261,
  STR    = 262,
	EOL    = 263,
  QUOTE  = 264
};

struct token *make_token(enum yytokentype, int, char*);
int   yylval;
char *yystr;
%}

%%
"("	         { return LPAR; }
")"	         { return RPAR; }
[0-9]+	     { yylval = atoi(yytext); return NUM; }
\n           { return EOL; }
[ \t\n]      { /* ignore white space */ }
\".*\"	     { return STR; }
[a-zA-Z][a-zA-Z0-9]* { return SYM; }
.            { err(1, "invalid symbol: %s\n", yytext); }
%%

/*
 * Goal: write LISP read-eval-print Loop and support
 *   @       Function
 *   0,1     boolean        (nand 0 1)    -> boolean
 *                          (eq 0 0)      -> boolean
 *   "str"   Strings        (streq str1 str2) -> boolean
 *   abc     Symboles       (symeq s1 s2)     -> boolean
 *   1234    Number/Integer [add,sub,div,mul,mod] -> num
 *                          [lt, numeq]    -> boolean
 *                          [num2bool]     -> boolean
 *
 *     (def <var> <symbol,number,#function)
 *     (lambda (<symbol> ..) (<sym..> ))
 *     (if <bool> <body> <else>)
 *     (quote a) -> a
 */

void
debug_token(struct token *x, int indent) {
  for (int i = 0; i < indent; i++) putc(' ', stderr);
  fprintf(stderr, "%d: (%p)", x->type, x);
  if (x->type == NUM) fprintf(stderr, "%d @ %p", x->v.num, x);
  if (x->type == STR || x->type == SYM)
    fprintf(stderr, "%s", x->v.str);
  putc('\n', stderr);
}

void
debug_ast(struct ast *a, int indent) {
  while (a) {
    for (int i = 0; i < indent; i++) putc(' ', stderr);
    fprintf(stderr, "AST: %s,a=%p,list=%p,token=%p,next=%p", (a->type == AST_TOK) ? "AST_TOK" :
            "AST_LIST", a, a->v.list, a->v.token, a->next);
    putc('\n', stderr);
    if (a->type == AST_LIST) {
      fprintf(stderr, "enter %p\n", a->v.list);
      fprintf(stderr, "==\n");
      debug_ast(a->v.list, indent+4);
      fprintf(stderr, "==\n");
    }
    else if(a->type == AST_TOK)
      debug_token(a->v.token, indent);
    a = a->next;
  }
}

struct token*
make_token(enum yytokentype type, int num, char* str)
{
  struct token *t = 
      (struct token*) calloc(1, sizeof(struct token));
	if (t == NULL)
    err(1, "malloc failed");
  t->type = type;
	if (type == NUM)
    t->v.num = num;
	else if (type == SYM || type == STR)
    t->v.str = strdup(str);
  return t;
}

struct ast*
make_ast(enum asttype type, struct token *t)
{
  struct ast *a = 
    (struct ast*) calloc(1, sizeof(struct ast));
  if (a == NULL)
    err(1, "malloc failed");
  a->type = type;
  if (type == AST_TOK)
    a->v.token = t;
  /* else if (type == AST_LIST) */
  /*   a->v.list = n;  */
  return a;
}

struct env*
make_env(char *name, struct ast *a)
{
   struct env *e = 
    (struct env*) calloc(1, sizeof(struct env));
   if (e == NULL)
     err(1, "malloc failed");
   e->name = name;
   e->a = a;
   return e;
}

/* upon parsing a list of tokens like:
 *    "(" -> "def" -> "(" -> "b" -> ")" -> "c" -> ")"
 * care must be taken when entering a new sub-list (b),
 * because the processing in the upper level must continue 
 * where the sub-list ended, i.e. "c" must be the next 
 * token on this level.
 */
struct token *next;
struct ast*
parse(struct token *t)
{
  if (t == NULL)
    err(1, "Unexpected token: <NULL>");

 if (t->type == LPAR) {
    struct ast *head = make_ast(AST_LIST, NULL);
    struct ast *a, *p;
    t = t->next;
    while (t->type != RPAR) {
      assert (t != NULL);
      a = parse(t);
      if (a == NULL)
        err(1, "syntax error");
      if (head->v.list == NULL)
        p = head->v.list = a;
      else
        p->next = a;
      p = a;
      t = next; /* skip all tokens that are processed */
      assert(t != NULL);
    }
    assert(t->type == RPAR);
    t = t->next;
    next = t;
    return head;
  } else if (t->type == NUM) {
    next = t->next;
    return make_ast(AST_TOK, t);
  } else if (t->type == SYM) {
    next = t->next;
    return make_ast(AST_TOK, t);
  } else if (t->type == STR) {
    next = t->next;
    return make_ast(AST_TOK, t);
  }
  return NULL;
}

void
append(struct env *e, struct env *a)
{
  if (e->next == NULL) {
    e->next = a;
    return;
  } else 
    append(e->next, a);
}

struct ast *
lookup(struct env *e, char *name)
{
  if (e == NULL)
    return NULL;
  for (struct env *t = e; t != NULL; t = t->next)
    if (strcmp(t->name, name) == 0)
      return t->a;
  return lookup(e->parent, name);
}

struct ast *
eval(struct ast *a, struct env *e)
{
  if (a == NULL) return NULL;

  if (a->type == AST_TOK) {
    if (a->v.token->type == NUM) {
      return make_ast(AST_TOK, a->v.token);
    }
    else if (a->v.token->type == SYM) {
      fprintf(stderr, "process: %s\n", a->v.token->v.str);
      if (strcmp(a->v.token->v.str, "def") == 0) {
        a = a->next;
        assert((a->next->type == AST_TOK && 
                a->next->v.token->type== STR)  ||
               (a->next->type == AST_TOK && 
                a->next->v.token->type == NUM) ||
               (a->next->type == AST_LIST)     ||      
               (a->next->type == AST_TOK && 
                a->next->v.token->type== SYM));
        append(e, make_env(a->v.token->v.str, eval(a->next, e)));
        return NULL;
      } else if (strcmp(a->v.token->v.str, "lm") == 0) {
      } else if (strcmp(a->v.token->v.str, "if") == 0) {
      } else if (strcmp(a->v.token->v.str, "add") == 0) {
        assert (a != NULL && 
                a->next != NULL &&
                a->next->next != NULL &&
                a->next->next->next == NULL);

        struct ast *op1 = eval(a->next, e);
        struct ast *op2 = eval(a->next->next, e);
        
        assert (op1 != NULL &&
                op1->type == AST_TOK &&
                op1->v.token->type == NUM &&
                op2 != NULL &&
                op2->type == AST_TOK &&
                op2->v.token->type == NUM);
        return make_ast(AST_TOK, 
                        make_token(NUM, 
                                   op1->v.token->v.num + 
                                     op2->v.token->v.num,
                                   NULL));
      } else if (strcmp(a->v.token->v.str, "q") == 0) {
        //return make_ast(AST_QUOTE, a->next);
        return a->next;
      } else {
        // fprintf(stderr, "Lookup: %s\n", a->v.token->v.str);
        struct ast *var = lookup(e, a->v.token->v.str);
        assert(var != NULL);            /* variable found? */
        return var;
      }
      /* XXX: nand, eq, streq, symeq, add, sub, div, mul, 
       *      mod, lt, numeq, num2bool
       */

    } else if (a->v.token->type == STR)
      return make_ast(AST_TOK, a->v.token);
  } else if (a->type == AST_LIST) {
    return eval(a->v.list, e);
  } else {
    fprintf(stderr, "%d\n", a->type);
    assert(0);
  }
  return NULL;
}

void
pr(struct ast *a)
{
  if (a == NULL) {
    fprintf(stderr, "null %p\n", a);
    return;
  }
  switch (a->type) {
  case AST_TOK:
    if (a->v.token->type == NUM)
      printf("%d\n", a->v.token->v.num);
    else if (a->v.token->type == STR)
      printf("%s\n", a->v.token->v.str);
    else if (a->v.token->type == SYM)
      printf("%s\n", a->v.token->v.str);
  }
}

int main(void)
{
  int tok;
  struct token *t, *u = NULL, *start = NULL;
  struct env default_env = { 0 }, *env;
  env = &default_env;
  env->name = "";

  while((tok = yylex())) {
    /* read token and create list */
    t = make_token(tok, yylval, yytext);
    if (start == NULL) start = t;
    if (u == NULL)     u = t;
    else               u->next = t;
    u = t;
    
    /* if EOL is reached, parse and evaluate form: */
    if (t->type == EOL) {
      struct token *x = start;
      // while (x != NULL) {debug_token(x, 0); x = x->next;}
     
      /* struct ast *tl = parse(start); */
      /* eval(tl); */
      pr(eval(parse(start), env));
      //debug_ast(tl, 1);
      
      /* start anew: */
      start = NULL;
    }
  }
}


/*
 * Local Variables:
 * mode: c;
 * eval: (message "main()")
 * fill-column: 80
 * comment-column: 40
 * indent-tabs-mode: nil
 * tab-width: 2
 * c-basic-offset: 2
 * End:
 */
